package com.bergerkiller.bukkit.nolagg;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.craftbukkit.CraftWorld;

import com.bergerkiller.bukkit.nolagg.bufferedtag.NBTBaseBuffered;
import com.bergerkiller.bukkit.nolagg.scheduler.ChunkOperation.Type;
import com.bergerkiller.bukkit.nolagg.scheduler.ChunkScheduler;

import net.minecraft.server.*;

@SuppressWarnings("rawtypes")
public class CustomChunkLoader extends ChunkRegionLoader implements IChunkLoader {

	public static void swapAll() {
		for (org.bukkit.World world : Bukkit.getServer().getWorlds()) {
			swap(world);
		}
	}
	public static void swap(org.bukkit.World world) {
		swap(((CraftWorld) world).getHandle());
	}
	public static void swap(WorldServer world) {
		swap(world.chunkProviderServer);
	}
	public static void swap(ChunkProviderServer cps) {
		try {
			Field e = ChunkProviderServer.class.getDeclaredField("e");
			e.setAccessible(true);
			IChunkLoader loader = (IChunkLoader) e.get(cps);
			Field a = ChunkRegionLoader.class.getDeclaredField("a");
			Field b = ChunkRegionLoader.class.getDeclaredField("b");
			Field c = ChunkRegionLoader.class.getDeclaredField("c");
			Field d = ChunkRegionLoader.class.getDeclaredField("d");
			a.setAccessible(true);
			b.setAccessible(true);
			c.setAccessible(true);
			d.setAccessible(true);
			List vara = (List) a.get(loader);
			Set varb = (Set) b.get(loader);
			Object varc = (Object) c.get(loader);
			File vard = (File) d.get(loader);
			e.set(cps, new CustomChunkLoader(vara, varb, varc, vard));
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to bind world to low-mem chunk loader:");
			t.printStackTrace();
		}
	}

	private List a = new ArrayList(); //contains the chunks remaining for saving
	private Set b = new HashSet();
	private Object c = new Object();
	private final File d;

	public CustomChunkLoader(List a, Set b, Object c, File d) {
		super(d);
		this.a = a;
		this.b = b;
		this.c = c;
		this.d = d;
	}

	private static Field pendinga;
	private static Field pendingb;
	public static ChunkCoordIntPair getPendingA(Object from) {
		try {
			if (pendinga == null) {
				pendinga = from.getClass().getDeclaredField("a");
				pendinga.setAccessible(true);
			}
			return (ChunkCoordIntPair) pendinga.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}
	public static NBTTagCompound getPendingB(Object from) {
		try {
			if (pendingb == null) {
				pendingb = from.getClass().getDeclaredField("b");
				pendingb.setAccessible(true);
			}
			return (NBTTagCompound) pendingb.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}
	
	/*
	 * This is called when a chunk has to be saved
	 */
	@Override
	public void a(World world, final Chunk chunk) {
		if (ChunkScheduler.isEnabled()) {
			try {
				if (chunk == latestUnloadedChunk) {
					ChunkScheduler.schedule(chunk, Type.UNLOAD);
				} else {
					ChunkScheduler.schedule(chunk, Type.SAVE);
				}
				latestUnloadedChunk = null;
				return;
			} catch (Throwable t) {
				NoLagg.log(Level.SEVERE, "Failed to unload chunk [" + chunk.x + "/" + chunk.z + "/" + chunk.world.getWorld().getName() + "]:");
				t.printStackTrace();
			}
		}
		latestUnloadedChunk = null;
		super.a(world, chunk);
	}
	
	/*
	 * This is called when a chunk has to be loaded from file
	 */
	@Override
	public Chunk a(World world, int i, int j) {
		if (!NoLagg.useBufferedChunkLoading) {
			return super.a(world, i, j);
		}
		NBTTagCompound nbttagcompound = null;
		ChunkCoordIntPair chunkcoordintpair = new ChunkCoordIntPair(i, j);
		synchronized (this.c) {
			if (this.b.contains(chunkcoordintpair)) {
				Object obj;
				for (int k = 0; k < this.a.size(); ++k) {
					obj = this.a.get(k);
					if (getPendingA(obj).equals(chunkcoordintpair)) {
						nbttagcompound = getPendingB(obj);
						break;
					}
				}
			}
		}

		if (nbttagcompound == null) {
			DataInputStream datainputstream = RegionFileCache.b(this.d, i, j);
			if (datainputstream == null) return null;

			//custom memory-efficient loading system
			BufferedChunk chunk = new BufferedChunk(world, i, j);
			if (loadFromStream(chunk, datainputstream)) {
				return chunk;
			} else {
				chunk.setFree();
			}

			//failure, wrong coords? Use native coding instead
			datainputstream = RegionFileCache.b(this.d, i, j);
			if (datainputstream == null) return null;
			nbttagcompound = NBTCompressedStreamTools.a((DataInput) datainputstream);
		}

		if (!nbttagcompound.hasKey("Level")) {
			System.out.println("Chunk file at " + i + "," + j + " is missing level data, skipping");
			return null;
		} else if (!nbttagcompound.getCompound("Level").hasKey("Blocks")) {
			System.out.println("Chunk file at " + i + "," + j + " is missing block data, skipping");
			return null;
		} else {
			Chunk chunk = ChunkLoader.a(world, nbttagcompound.getCompound("Level"));

			if (!chunk.a(i, j)) {
				System.out.println("Chunk file at " + i + "," + j + " is in the wrong location; relocating. (Expected " + i + ", " + j + ", got " + chunk.x + ", " + chunk.z + ")");
				nbttagcompound.setInt("xPos", i);
				nbttagcompound.setInt("zPos", j);
				chunk = ChunkLoader.a(world, nbttagcompound.getCompound("Level"));
			}

			chunk.h();
			return chunk;
		}
	}

	public static boolean loadFromStream(final BufferedChunk chunk, DataInputStream stream) {
		try {
			byte type = stream.readByte();
			if (type == 0) throw new IOException("Root tag must be a named compound tag");
			String name = stream.readUTF();
			//BufferedTag reading with buffers
			final Map<String, byte[]> buffers = new HashMap<String, byte[]>();
			buffers.put("Blocks", chunk.b);
			buffers.put("Data", chunk.g.a);
			buffers.put("SkyLight", chunk.h.a);
			buffers.put("BlockLight", chunk.i.a);
			buffers.put("HeightMap", chunk.heightMap);
			NBTTagCompound comp = (NBTTagCompound) NBTBaseBuffered.readFrom(type, name, buffers, stream);
			comp = comp.getCompound("Level");
			if (comp == null) throw new Exception("Chunk at " + chunk.x + "," + chunk.z + " is missing level data");
			if (!comp.hasKey("Blocks")) throw new Exception("Chunk at " + chunk.x + "," + chunk.z + " is missing block data");
			return load(chunk, comp);
		} catch (Exception ex) {
			NoLagg.log(Level.SEVERE, "An error occurred while loading a chunk:");
			ex.printStackTrace();
			NoLagg.log(Level.WARNING, "Native methods will now try to load it...");
			return false;
		}
	}
	
	public static NibbleArray getNibble(final NibbleArray old, final byte[] newdata, final int bits) {
		if (old.a.length != newdata.length) {
			return new NibbleArray(newdata, bits);
		} else {
			return old;
		}
	}
	
	/*
	 * Now uses custom saving thread
	 */
	private static Chunk latestUnloadedChunk = null;
	public static void expectChunkUnload(org.bukkit.Chunk chunk) {
		expectChunkUnload(ChunkHandler.getNative(chunk));
	}
	public static void expectChunkUnload(Chunk chunk) {
		latestUnloadedChunk = chunk;
	}
	
	public void saveDirect(World world, Chunk chunk) {
	    //generate tag compound
        NBTTagCompound nbttagcompound = new NBTTagCompound();
        NBTTagCompound nbttagcompound1 = new NBTTagCompound();
        nbttagcompound.set("Level", nbttagcompound1);
        ChunkLoader.a(chunk, world, nbttagcompound1);
	
        try {
            DataOutputStream dataoutputstream = RegionFileCache.c(this.d, chunk.x, chunk.z);
            NBTCompressedStreamTools.a(nbttagcompound, (DataOutput) dataoutputstream);
            dataoutputstream.flush();
            dataoutputstream.close();
        } catch (Throwable t) {
        	t.printStackTrace();
        }
	}
	
	public static boolean load(BufferedChunk chunk, NBTTagCompound nbttagcompound) {
		int i = nbttagcompound.getInt("xPos");
		int j = nbttagcompound.getInt("zPos");
		if (chunk.x != i || chunk.z != j) return false;

		chunk.b = nbttagcompound.getByteArray("Blocks");
		chunk.g = getNibble(chunk.g, nbttagcompound.getByteArray("Data"), chunk.world.heightBits);
		chunk.h = getNibble(chunk.h, nbttagcompound.getByteArray("SkyLight"), chunk.world.heightBits);
		chunk.i = getNibble(chunk.i, nbttagcompound.getByteArray("BlockLight"), chunk.world.heightBits);
		
		chunk.heightMap = nbttagcompound.getByteArray("HeightMap");
		chunk.done = nbttagcompound.getBoolean("TerrainPopulated");
		if (!chunk.g.a()) {
			chunk.g = new NibbleArray(chunk.b.length, chunk.world.heightBits);
		}

		if (chunk.heightMap == null || !chunk.h.a()) {
			chunk.heightMap = new byte[256];
			chunk.h = new NibbleArray(chunk.b.length, chunk.world.heightBits);
			chunk.initLighting();
		}

		if (!chunk.i.a()) {
			chunk.i = new NibbleArray(chunk.b.length, chunk.world.heightBits);
			chunk.a();
		}

		NBTTagList nbttaglist = nbttagcompound.getList("Entities");

		if (nbttaglist != null) {
			for (int k = 0; k < nbttaglist.size(); ++k) {
				NBTTagCompound nbttagcompound1 = (NBTTagCompound) nbttaglist.get(k);
				Entity entity = EntityTypes.a(nbttagcompound1, chunk.world);

				chunk.s = true;
				if (entity != null) {
					chunk.a(entity);
				}
			}
		}

		NBTTagList nbttaglist1 = nbttagcompound.getList("TileEntities");

		if (nbttaglist1 != null) {
			for (int l = 0; l < nbttaglist1.size(); ++l) {
				NBTTagCompound nbttagcompound2 = (NBTTagCompound) nbttaglist1.get(l);
				TileEntity tileentity = TileEntity.c(nbttagcompound2);

				if (tileentity != null) {
					chunk.a(tileentity);
				}
			}
		}

		if (nbttagcompound.hasKey("TileTicks")) {
			NBTTagList nbttaglist2 = nbttagcompound.getList("TileTicks");

			if (nbttaglist2 != null) {
				for (int i1 = 0; i1 < nbttaglist2.size(); ++i1) {
					NBTTagCompound nbttagcompound3 = (NBTTagCompound) nbttaglist2.get(i1);

					chunk.world.d(nbttagcompound3.getInt("x"), nbttagcompound3.getInt("y"), nbttagcompound3.getInt("z"), nbttagcompound3.getInt("i"), nbttagcompound3.getInt("t"));
				}
			}
		}

		return true;
	}

}
